@@ -20,7 +20,7 @@
 package syscalls
 
 // Generated using arch-syscall-dump test tool from libseccomp tree, git
-// revision 584ca7a5e69d87a4c2c4e4c07ce8415fa59e1351.
+// revision bf747eb21e428c2b3ead6ebcca27951b681963a0.
 var SeccompSyscalls = []string{
 	"_llseek",
 	"_newselect",
@@ -52,11 +52,17 @@ var SeccompSyscalls = []string{
 	"chown32",
 	"chroot",
 	"clock_adjtime",
+	"clock_adjtime64",
 	"clock_getres",
+	"clock_getres_time64",
 	"clock_gettime",
+	"clock_gettime64",
 	"clock_nanosleep",
+	"clock_nanosleep_time64",
 	"clock_settime",
+	"clock_settime64",
 	"clone",
+	"clone3",
 	"close",
 	"connect",
 	"copy_file_range",
@@ -100,7 +106,11 @@ var SeccompSyscalls = []string{
 	"flock",
 	"fork",
 	"fremovexattr",
+	"fsconfig",
 	"fsetxattr",
+	"fsmount",
+	"fsopen",
+	"fspick",
 	"fstat",
 	"fstat64",
 	"fstatat64",
@@ -111,6 +121,7 @@ var SeccompSyscalls = []string{
 	"ftruncate",
 	"ftruncate64",
 	"futex",
+	"futex_time64",
 	"futimesat",
 	"get_kernel_syms",
 	"get_mempolicy",
@@ -163,6 +174,7 @@ var SeccompSyscalls = []string{
 	"io_destroy",
 	"io_getevents",
 	"io_pgetevents",
+	"io_pgetevents_time64",
 	"io_setup",
 	"io_submit",
 	"io_uring_enter",
@@ -211,6 +223,7 @@ var SeccompSyscalls = []string{
 	"mmap2",
 	"modify_ldt",
 	"mount",
+	"move_mount",
 	"move_pages",
 	"mprotect",
 	"mpx",
@@ -218,7 +231,9 @@ var SeccompSyscalls = []string{
 	"mq_notify",
 	"mq_open",
 	"mq_timedreceive",
+	"mq_timedreceive_time64",
 	"mq_timedsend",
+	"mq_timedsend_time64",
 	"mq_unlink",
 	"mremap",
 	"msgctl",
@@ -243,6 +258,7 @@ var SeccompSyscalls = []string{
 	"oldwait4",
 	"open",
 	"open_by_handle_at",
+	"open_tree",
 	"openat",
 	"pause",
 	"pciconfig_iobase",
@@ -250,6 +266,8 @@ var SeccompSyscalls = []string{
 	"pciconfig_write",
 	"perf_event_open",
 	"personality",
+	"pidfd_open",
+	"pidfd_send_signal",
 	"pipe",
 	"pipe2",
 	"pivot_root",
@@ -258,6 +276,7 @@ var SeccompSyscalls = []string{
 	"pkey_mprotect",
 	"poll",
 	"ppoll",
+	"ppoll_time64",
 	"prctl",
 	"pread64",
 	"preadv",
@@ -268,6 +287,7 @@ var SeccompSyscalls = []string{
 	"prof",
 	"profil",
 	"pselect6",
+	"pselect6_time64",
 	"ptrace",
 	"putpmsg",
 	"pwrite64",
@@ -285,6 +305,7 @@ var SeccompSyscalls = []string{
 	"recv",
 	"recvfrom",
 	"recvmmsg",
+	"recvmmsg_time64",
 	"recvmsg",
 	"remap_file_pages",
 	"removexattr",
@@ -302,6 +323,7 @@ var SeccompSyscalls = []string{
 	"rt_sigreturn",
 	"rt_sigsuspend",
 	"rt_sigtimedwait",
+	"rt_sigtimedwait_time64",
 	"rt_tgsigqueueinfo",
 	"rtas",
 	"s390_guarded_storage",
@@ -316,6 +338,7 @@ var SeccompSyscalls = []string{
 	"sched_getparam",
 	"sched_getscheduler",
 	"sched_rr_get_interval",
+	"sched_rr_get_interval_time64",
 	"sched_setaffinity",
 	"sched_setattr",
 	"sched_setparam",
@@ -328,6 +351,7 @@ var SeccompSyscalls = []string{
 	"semget",
 	"semop",
 	"semtimedop",
+	"semtimedop_time64",
 	"send",
 	"sendfile",
 	"sendfile64",
@@ -421,11 +445,15 @@ var SeccompSyscalls = []string{
 	"timer_delete",
 	"timer_getoverrun",
 	"timer_gettime",
+	"timer_gettime64",
 	"timer_settime",
+	"timer_settime64",
 	"timerfd",
 	"timerfd_create",
 	"timerfd_gettime",
+	"timerfd_gettime64",
 	"timerfd_settime",
+	"timerfd_settime64",
 	"times",
 	"tkill",
 	"truncate",
@@ -447,6 +475,7 @@ var SeccompSyscalls = []string{
 	"ustat",
 	"utime",
 	"utimensat",
+	"utimensat_time64",
 	"utimes",
 	"vfork",
 	"vhangup",
