@@ -34,7 +34,7 @@ import (
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil"
-	"github.com/snapcore/snapd/selinux"
+	"github.com/snapcore/snapd/sandbox/selinux"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/snap/snaptest"
 	"github.com/snapcore/snapd/testutil"
@@ -50,15 +50,30 @@ hooks:
  configure:
 `)
 
+var mockYamlBaseNone1 = []byte(`name: snapname1
+version: 1.0
+base: none
+apps:
+ app:
+  command: run-app
+`)
+
+var mockYamlBaseNone2 = []byte(`name: snapname2
+version: 1.0
+base: none
+hooks:
+ configure:
+`)
+
 type RunSuite struct {
 	fakeHome string
-	SnapSuite
+	BaseSnapSuite
 }
 
 var _ = check.Suite(&RunSuite{})
 
 func (s *RunSuite) SetUpTest(c *check.C) {
-	s.SnapSuite.SetUpTest(c)
+	s.BaseSnapSuite.SetUpTest(c)
 	s.fakeHome = c.MkDir()
 
 	u, err := user.Current()
@@ -94,6 +109,24 @@ func (s *RunSuite) TestInvalidParameters(c *check.C) {
 	c.Check(err, check.ErrorMatches, ".*too many arguments for hook \"configure\": bar.*")
 }
 
+func (s *RunSuite) TestRunCmdWithBaseNone(c *check.C) {
+	defer mockSnapConfine(dirs.DistroLibExecDir)()
+
+	// mock installed snap
+	snaptest.MockSnapCurrent(c, string(mockYamlBaseNone1), &snap.SideInfo{
+		Revision: snap.R("1"),
+	})
+	snaptest.MockSnapCurrent(c, string(mockYamlBaseNone2), &snap.SideInfo{
+		Revision: snap.R("1"),
+	})
+
+	_, err := snaprun.Parser(snaprun.Client()).ParseArgs([]string{"run", "--", "snapname1.app", "--arg1", "arg2"})
+	c.Assert(err, check.ErrorMatches, `cannot run hooks / applications with base \"none\"`)
+
+	_, err = snaprun.Parser(snaprun.Client()).ParseArgs([]string{"run", "--hook=configure", "--", "snapname2"})
+	c.Assert(err, check.ErrorMatches, `cannot run hooks / applications with base \"none\"`)
+}
+
 func (s *RunSuite) TestSnapRunWhenMissingConfine(c *check.C) {
 	_, r := logger.MockLogger()
 	defer r()
