@@ -28,7 +28,7 @@ import (
 
 	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/image"
-	"github.com/snapcore/snapd/selinux"
+	"github.com/snapcore/snapd/sandbox/selinux"
 	"github.com/snapcore/snapd/store"
 )
 
@@ -39,16 +39,18 @@ var (
 
 	FirstNonOptionIsRun = firstNonOptionIsRun
 
-	CreateUserDataDirs = createUserDataDirs
-	ResolveApp         = resolveApp
-	SnapdHelperPath    = snapdHelperPath
-	SortByPath         = sortByPath
-	AdviseCommand      = adviseCommand
-	Antialias          = antialias
-	FormatChannel      = fmtChannel
-	PrintDescr         = printDescr
-	WrapFlow           = wrapFlow
-	TrueishJSON        = trueishJSON
+	CreateUserDataDirs  = createUserDataDirs
+	ResolveApp          = resolveApp
+	SnapdHelperPath     = snapdHelperPath
+	SortByPath          = sortByPath
+	AdviseCommand       = adviseCommand
+	Antialias           = antialias
+	FormatChannel       = fmtChannel
+	PrintDescr          = printDescr
+	WrapFlow            = wrapFlow
+	TrueishJSON         = trueishJSON
+	CompletionHandler   = completionHandler
+	MarkForNoCompletion = markForNoCompletion
 
 	CanUnicode           = canUnicode
 	ColorTable           = colorTable
@@ -80,8 +82,22 @@ var (
 	InterfacesDeprecationNotice = interfacesDeprecationNotice
 
 	SignalNotify = signalNotify
+
+	SortTimingsTasks = sortTimingsTasks
+
+	PrintInstallHint = printInstallHint
 )
 
+func HiddenCmd(descr string, completeHidden bool) *cmdInfo {
+	return &cmdInfo{
+		shortHelp:      descr,
+		hidden:         true,
+		completeHidden: completeHidden,
+	}
+}
+
+type ChangeTimings = changeTimings
+
 func NewInfoWriter(w writeflusher) *infoWriter {
 	return &infoWriter{
 		writeflusher: w,
