@@ -26,7 +26,6 @@ import (
 	"github.com/snapcore/snapd/interfaces/mount"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/release"
-	"github.com/snapcore/snapd/snap"
 )
 
 const desktopSummary = `allows access to basic graphical desktop resources`
@@ -105,7 +104,7 @@ dbus (receive)
     bus=session
     path=/org/freedesktop/Notifications
     interface=org.freedesktop.Notifications
-    member={ActionInvoked,NotificationClosed}
+    member={ActionInvoked,NotificationClosed,NotificationReplied}
     peer=(label=unconfined),
 
 # DesktopAppInfo Launched
@@ -198,25 +197,14 @@ dbus (receive, send)
     interface=org.freedesktop.DBus.Properties
     path=/org/freedesktop/portal/{desktop,documents}{,/**}
     peer=(label=unconfined),
+
+# These accesses are noisy and applications can't do anything with the found
+# icon files, so explicitly deny to silence the denials
+deny /var/lib/snapd/desktop/icons/ r,
 `
 
-type desktopInterface struct{}
-
-func (iface *desktopInterface) Name() string {
-	return "desktop"
-}
-
-func (iface *desktopInterface) StaticInfo() interfaces.StaticInfo {
-	return interfaces.StaticInfo{
-		Summary:              desktopSummary,
-		ImplicitOnClassic:    true,
-		BaseDeclarationSlots: desktopBaseDeclarationSlots,
-	}
-}
-
-func (iface *desktopInterface) AutoConnect(*snap.PlugInfo, *snap.SlotInfo) bool {
-	// allow what declarations allowed
-	return true
+type desktopInterface struct {
+	commonInterface
 }
 
 func (iface *desktopInterface) fontconfigDirs() []string {
@@ -231,7 +219,7 @@ func (iface *desktopInterface) AppArmorConnectedPlug(spec *apparmor.Specificatio
 	spec.AddSnippet(desktopConnectedPlugAppArmor)
 
 	// Allow mounting document portal
-	emit := spec.EmitUpdateNSFunc()
+	emit := spec.AddUpdateNSf
 	emit("  # Mount the document portal\n")
 	emit("  mount options=(bind) /run/user/[0-9]*/doc/by-app/snap.%s/ -> /run/user/[0-9]*/doc/,\n", plug.Snap().InstanceName())
 	emit("  umount /run/user/[0-9]*/doc/,\n\n")
@@ -286,5 +274,12 @@ func (iface *desktopInterface) MountConnectedPlug(spec *mount.Specification, plu
 }
 
 func init() {
-	registerIface(&desktopInterface{})
+	registerIface(&desktopInterface{
+		commonInterface: commonInterface{
+			name:                 "desktop",
+			summary:              desktopSummary,
+			implicitOnClassic:    true,
+			baseDeclarationSlots: desktopBaseDeclarationSlots,
+		},
+	})
 }
