@@ -27,6 +27,7 @@ import (
 	"github.com/snapcore/snapd/interfaces/seccomp"
 	"github.com/snapcore/snapd/interfaces/udev"
 	"github.com/snapcore/snapd/release"
+	apparmor_sandbox "github.com/snapcore/snapd/sandbox/apparmor"
 	"github.com/snapcore/snapd/snap"
 )
 
@@ -82,8 +83,7 @@ const dockerSupportConnectedPlugAppArmor = `
 /run/ipam-state/k8s-*/lock k,
 
 # Socket for docker-container-shim
-unix (bind,listen) type=stream addr="@/containerd-shim/moby/*/shim.sock\x00",
-unix (bind,listen) type=stream addr="@/containerd-shim/k8s.io/*/shim.sock\x00",
+unix (bind,listen) type=stream addr="@/containerd-shim/**/shim.sock\x00",
 
 /{,var/}run/mount/utab r,
 
@@ -95,6 +95,7 @@ unix (bind,listen) type=stream addr="@/containerd-shim/k8s.io/*/shim.sock\x00",
 
 # Limited read access to specific bits of /sys
 /sys/kernel/mm/hugepages/ r,
+/sys/kernel/mm/transparent_hugepage/{,**} r,
 /sys/fs/cgroup/cpuset/cpuset.cpus r,
 /sys/fs/cgroup/cpuset/cpuset.mems r,
 /sys/module/apparmor/parameters/enabled r,
@@ -626,7 +627,7 @@ func (iface *dockerSupportInterface) StaticInfo() interfaces.StaticInfo {
 }
 
 var (
-	parserFeatures = release.AppArmorParserFeatures
+	parserFeatures = apparmor_sandbox.ParserFeatures
 )
 
 func (iface *dockerSupportInterface) UDevConnectedPlug(spec *udev.Specification, plug *interfaces.ConnectedPlug, slot *interfaces.ConnectedSlot) error {
