@@ -386,7 +386,7 @@ func New(cfg *Config, dauthCtx DeviceAndAuthContext) *Store {
 
 // API endpoint paths
 const (
-	// see https://wiki.ubuntu.com/AppStore/Interfaces/ClickPackageIndex
+	// see https://dashboard.snapcraft.io/docs/
 	// XXX: Repeating "api/" here is cumbersome, but the next generation
 	// of store APIs will probably drop that prefix (since it now
 	// duplicates the hostname), and we may want to switch to v2 APIs
@@ -929,6 +929,7 @@ func (s *Store) newRequest(ctx context.Context, reqOptions *requestOptions, user
 	req.Header.Set(hdrSnapDeviceArchitecture[reqOptions.APILevel], s.architecture)
 	req.Header.Set(hdrSnapDeviceSeries[reqOptions.APILevel], s.series)
 	req.Header.Set(hdrSnapClassic[reqOptions.APILevel], strconv.FormatBool(release.OnClassic))
+	req.Header.Set("Snap-Device-Capabilities", "default-tracks")
 	if cua := ClientUserAgent(ctx); cua != "" {
 		req.Header.Set("Snap-Client-User-Agent", cua)
 	}
@@ -1528,12 +1529,19 @@ func downloadImpl(ctx context.Context, name, sha3_384, downloadURL string, user
 			return fmt.Errorf("The download has been cancelled: %s", ctx.Err())
 		}
 		if finalErr != nil {
-			if httputil.ShouldRetryError(attempt, finalErr) {
+			if httputil.ShouldRetryAttempt(attempt, finalErr) {
 				continue
 			}
 			break
 		}
-
+		if resume > 0 && resp.StatusCode != 206 {
+			logger.Debugf("server does not support resume")
+			if _, err := w.Seek(0, os.SEEK_SET); err != nil {
+				return err
+			}
+			h = crypto.SHA3_384.New()
+			resume = 0
+		}
 		if httputil.ShouldRetryHttpResponse(attempt, resp) {
 			resp.Body.Close()
 			continue
@@ -1565,7 +1573,7 @@ func downloadImpl(ctx context.Context, name, sha3_384, downloadURL string, user
 		_, finalErr = io.Copy(mw, limiter)
 		pbar.Finished()
 		if finalErr != nil {
-			if httputil.ShouldRetryError(attempt, finalErr) {
+			if httputil.ShouldRetryAttempt(attempt, finalErr) {
 				// error while downloading should resume
 				var seekerr error
 				resume, seekerr = w.Seek(0, os.SEEK_END)
@@ -1605,19 +1613,27 @@ func downloadImpl(ctx context.Context, name, sha3_384, downloadURL string, user
 }
 
 // DownloadStream will copy the snap from the request to the io.Reader
-func (s *Store) DownloadStream(ctx context.Context, name string, downloadInfo *snap.DownloadInfo, user *auth.UserState) (io.ReadCloser, error) {
+func (s *Store) DownloadStream(ctx context.Context, name string, downloadInfo *snap.DownloadInfo, resume int64, user *auth.UserState) (io.ReadCloser, int, error) {
+	// XXX: coverage of this is rather poor
 	if path := s.cacher.GetPath(downloadInfo.Sha3_384); path != "" {
 		logger.Debugf("Cache hit for SHA3_384 …%.5s.", downloadInfo.Sha3_384)
 		file, err := os.OpenFile(path, os.O_RDONLY, 0600)
 		if err != nil {
-			return nil, err
+			return nil, 0, err
 		}
-		return file, nil
+		if resume == 0 {
+			return file, 200, nil
+		}
+		_, err = file.Seek(resume, os.SEEK_SET)
+		if err != nil {
+			return nil, 0, err
+		}
+		return file, 206, nil
 	}
 
 	authAvail, err := s.authAvailable(user)
 	if err != nil {
-		return nil, err
+		return nil, 0, err
 	}
 
 	downloadURL := downloadInfo.AnonDownloadURL
@@ -1627,25 +1643,28 @@ func (s *Store) DownloadStream(ctx context.Context, name string, downloadInfo *s
 
 	storeURL, err := url.Parse(downloadURL)
 	if err != nil {
-		return nil, err
+		return nil, 0, err
 	}
 
 	cdnHeader, err := s.cdnHeader()
 	if err != nil {
-		return nil, err
+		return nil, 0, err
 	}
 
-	resp, err := doDownloadReq(ctx, storeURL, cdnHeader, s, user)
+	resp, err := doDownloadReq(ctx, storeURL, cdnHeader, resume, s, user)
 	if err != nil {
-		return nil, err
+		return nil, 0, err
 	}
-	return resp.Body, nil
+	return resp.Body, resp.StatusCode, nil
 }
 
-var doDownloadReq = doDowloadReqImpl
+var doDownloadReq = doDownloadReqImpl
 
-func doDowloadReqImpl(ctx context.Context, storeURL *url.URL, cdnHeader string, s *Store, user *auth.UserState) (*http.Response, error) {
+func doDownloadReqImpl(ctx context.Context, storeURL *url.URL, cdnHeader string, resume int64, s *Store, user *auth.UserState) (*http.Response, error) {
 	reqOptions := downloadReqOpts(storeURL, cdnHeader, nil)
+	if resume > 0 {
+		reqOptions.ExtraHeaders["Range"] = fmt.Sprintf("bytes=%d-", resume)
+	}
 	return s.doRequest(ctx, httputil.NewHTTPClient(&httputil.ClientOptions{Proxy: s.proxy}), reqOptions, user)
 }
 
@@ -1676,13 +1695,10 @@ func (s *Store) downloadDelta(deltaName string, downloadInfo *snap.DownloadInfo,
 }
 
 func getXdelta3Cmd(args ...string) (*exec.Cmd, error) {
-	switch {
-	case osutil.ExecutableExists("xdelta3"):
+	if osutil.ExecutableExists("xdelta3") {
 		return exec.Command("xdelta3", args...), nil
-	case osutil.FileExists(filepath.Join(dirs.SnapMountDir, "/core/current/usr/bin/xdelta3")):
-		return cmdutil.CommandFromSystemSnap("/usr/bin/xdelta3", args...)
 	}
-	return nil, fmt.Errorf("cannot find xdelta3 binary in PATH or core snap")
+	return cmdutil.CommandFromSystemSnap("/usr/bin/xdelta3", args...)
 }
 
 // applyDelta generates a target snap from a previously downloaded snap and a downloaded delta.
@@ -2101,6 +2117,7 @@ type snapActionResult struct {
 	Name             string    `json:"name,omitempty"`
 	Snap             storeSnap `json:"snap"`
 	EffectiveChannel string    `json:"effective-channel,omitempty"`
+	RedirectChannel  string    `json:"redirect-channel,omitempty"`
 	Error            struct {
 		Code    string `json:"code"`
 		Message string `json:"message"`
@@ -2131,7 +2148,7 @@ var snapActionFields = jsonutil.StructFields((*storeSnap)(nil))
 // current installed snaps in currentSnaps. If the request was overall
 // successul (200) but there were reported errors it will return both
 // the snap infos and an SnapActionError.
-func (s *Store) SnapAction(ctx context.Context, currentSnaps []*CurrentSnap, actions []*SnapAction, user *auth.UserState, opts *RefreshOptions) ([]*snap.Info, error) {
+func (s *Store) SnapAction(ctx context.Context, currentSnaps []*CurrentSnap, actions []*SnapAction, user *auth.UserState, opts *RefreshOptions) ([]SnapActionResult, error) {
 	if opts == nil {
 		opts = &RefreshOptions{}
 	}
@@ -2143,7 +2160,7 @@ func (s *Store) SnapAction(ctx context.Context, currentSnaps []*CurrentSnap, act
 
 	authRefreshes := 0
 	for {
-		snaps, err := s.snapAction(ctx, currentSnaps, actions, user, opts)
+		sars, err := s.snapAction(ctx, currentSnaps, actions, user, opts)
 
 		if saErr, ok := err.(*SnapActionError); ok && authRefreshes < 2 && len(saErr.Other) > 0 {
 			// do we need to try to refresh auths?, 2 tries
@@ -2171,7 +2188,7 @@ func (s *Store) SnapAction(ctx context.Context, currentSnaps []*CurrentSnap, act
 			}
 		}
 
-		return snaps, err
+		return sars, err
 	}
 }
 
@@ -2196,7 +2213,14 @@ func genInstanceKey(curSnap *CurrentSnap, salt string) (string, error) {
 	return fmt.Sprintf("%s:%s", curSnap.SnapID, enc), nil
 }
 
-func (s *Store) snapAction(ctx context.Context, currentSnaps []*CurrentSnap, actions []*SnapAction, user *auth.UserState, opts *RefreshOptions) ([]*snap.Info, error) {
+// SnapActionResult encapsulates the non-error result of a single
+// action of the SnapAction call.
+type SnapActionResult struct {
+	*snap.Info
+	RedirectChannel string
+}
+
+func (s *Store) snapAction(ctx context.Context, currentSnaps []*CurrentSnap, actions []*SnapAction, user *auth.UserState, opts *RefreshOptions) ([]SnapActionResult, error) {
 
 	// TODO: the store already requires instance-key but doesn't
 	// yet support repeating in context or sending actions for the
@@ -2358,7 +2382,7 @@ func (s *Store) snapAction(ctx context.Context, currentSnaps []*CurrentSnap, act
 	downloadErrors := make(map[string]error)
 	var otherErrors []error
 
-	var snaps []*snap.Info
+	var sars []SnapActionResult
 	for _, res := range results.Results {
 		if res.Result == "error" {
 			if a := installs[res.InstanceKey]; a != nil {
@@ -2423,7 +2447,7 @@ func (s *Store) snapAction(ctx context.Context, currentSnaps []*CurrentSnap, act
 		_, instanceKey := snap.SplitInstanceName(instanceName)
 		snapInfo.InstanceKey = instanceKey
 
-		snaps = append(snaps, snapInfo)
+		sars = append(sars, SnapActionResult{Info: snapInfo, RedirectChannel: res.RedirectChannel})
 	}
 
 	for _, errObj := range results.ErrorList {
@@ -2441,7 +2465,7 @@ func (s *Store) snapAction(ctx context.Context, currentSnaps []*CurrentSnap, act
 		if len(downloadErrors) == 0 {
 			downloadErrors = nil
 		}
-		return snaps, &SnapActionError{
+		return sars, &SnapActionError{
 			NoResults: len(results.Results) == 0,
 			Refresh:   refreshErrors,
 			Install:   installErrors,
@@ -2450,7 +2474,7 @@ func (s *Store) snapAction(ctx context.Context, currentSnaps []*CurrentSnap, act
 		}
 	}
 
-	return snaps, nil
+	return sars, nil
 }
 
 // abbreviated info structs just for the download info
