@@ -270,9 +270,7 @@ prepare_project() {
 
     create_test_user
 
-    if ! distro_update_package_db; then
-        echo "Error updating the package db, continue with the system preparation"
-    fi
+    distro_update_package_db
 
     if [[ "$SPREAD_SYSTEM" == arch-* ]]; then
         # perform system upgrade on Arch so that we run with most recent kernel
@@ -363,6 +361,11 @@ prepare_project() {
         ubuntu-*)
             # Ubuntu is the only system where snapd is preinstalled
             distro_purge_package snapd
+            # XXX: the original package's purge may have left socket units behind
+            find /etc/systemd/system -name "snap.*.socket" | while read -r f; do
+                systemctl stop "$(basename "$f")" || true
+                rm -f "$f"
+            done
             ;;
         *)
             # snapd state directory must not exist when the package is not
@@ -371,9 +374,7 @@ prepare_project() {
             ;;
     esac
 
-    if ! install_pkg_dependencies; then
-        echo "Error installing test dependencies, continue with the system preparation"
-    fi
+    install_pkg_dependencies
 
     # We take a special case for Debian/Ubuntu where we install additional build deps
     # base on the packaging. In Fedora/Suse this is handled via mock/osc
@@ -446,18 +447,14 @@ prepare_project() {
     go get ./tests/lib/fakedevicesvc
     go get ./tests/lib/systemd-escape
 
-    # Disable journald rate limiting
-    mkdir -p /etc/systemd/journald.conf.d
-    # The RateLimitIntervalSec key is not supported on some systemd versions causing
-    # the journal rate limit could be considered as not valid and discarded in concecuence.
-    # RateLimitInterval key is supported in old systemd versions and in new ones as well,
-    # maintaining backward compatibility.
-    cat <<-EOF > /etc/systemd/journald.conf.d/no-rate-limit.conf
-    [Journal]
-    RateLimitInterval=0
-    RateLimitBurst=0
-EOF
-    systemctl restart systemd-journald.service
+    # On core systems, the journal service is configured once the final core system
+    # is created and booted what is done during the first test suite preparation
+    if is_classic_system; then
+        # shellcheck source=tests/lib/prepare.sh
+        . "$TESTSLIB"/prepare.sh
+        disable_journald_rate_limiting
+        disable_journald_start_limiting
+    fi
 }
 
 prepare_project_each() {
@@ -501,7 +498,12 @@ prepare_suite_each() {
     echo -n "$SPREAD_JOB " >> "$RUNTIME_STATE_PATH/runs"
     # shellcheck source=tests/lib/reset.sh
     "$TESTSLIB"/reset.sh --reuse-core
-    # Reset systemd journal cursor.
+    # Restart journal log and reset systemd journal cursor.
+    if ! systemctl restart systemd-journald.service; then
+        systemctl status systemd-journald.service || true
+        echo "Failed to restart systemd-journald.service, exiting..."
+        exit 1
+    fi
     start_new_journalctl_log
 
     echo "Install the snaps profiler snap"
