@@ -22,9 +22,9 @@ package wrappers
 import (
 	"bytes"
 	"fmt"
-	"math/rand"
 	"os"
 	"path/filepath"
+	"strconv"
 	"strings"
 	"text/template"
 	"time"
@@ -33,7 +33,10 @@ import (
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/osutil/sys"
+	"github.com/snapcore/snapd/randutil"
+	"github.com/snapcore/snapd/release"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/strutil"
 	"github.com/snapcore/snapd/systemd"
 	"github.com/snapcore/snapd/timeout"
 	"github.com/snapcore/snapd/timeutil"
@@ -199,9 +202,20 @@ func StartServices(apps []*snap.AppInfo, inter interacter, tm timings.Measurer)
 }
 
 // AddSnapServices adds service units for the applications from the snap which are services.
-func AddSnapServices(s *snap.Info, inter interacter) (err error) {
+func AddSnapServices(s *snap.Info, disabledSvcs []string, inter interacter) (err error) {
 	if s.GetType() == snap.TypeSnapd {
-		return writeSnapdServicesOnCore(s, inter)
+		return fmt.Errorf("internal error: adding explicit services for snapd snap is unexpected")
+	}
+
+	// check if any previously disabled services are now no longer services and
+	// log messages about that
+	for _, svc := range disabledSvcs {
+		app, ok := s.Apps[svc]
+		if !ok {
+			logger.Noticef("previously disabled service %s no longer exists", svc)
+		} else if !app.IsService() {
+			logger.Noticef("previously disabled service %s is now an app and not a service", svc)
+		}
 	}
 
 	sysd := systemd.New(dirs.GlobalRootDir, systemd.SystemMode, inter)
@@ -228,6 +242,9 @@ func AddSnapServices(s *snap.Info, inter interacter) (err error) {
 		}
 	}()
 
+	// TODO: remove once services get enabled on start and not when created.
+	preseedMode := release.PreseedMode
+
 	for _, app := range s.Apps {
 		if !app.IsService() {
 			continue
@@ -277,13 +294,21 @@ func AddSnapServices(s *snap.Info, inter interacter) (err error) {
 		}
 
 		svcName := app.ServiceName()
-		if err := sysd.Enable(svcName); err != nil {
-			return err
+
+		if strutil.ListContains(disabledSvcs, app.Name) {
+			// service is disabled, nothing to do
+			continue
+		}
+
+		if !preseedMode() {
+			if err := sysd.Enable(svcName); err != nil {
+				return err
+			}
 		}
 		enabled = append(enabled, svcName)
 	}
 
-	if len(written) > 0 {
+	if len(written) > 0 && !preseedMode() {
 		if err := sysd.DaemonReload(); err != nil {
 			return err
 		}
@@ -292,6 +317,22 @@ func AddSnapServices(s *snap.Info, inter interacter) (err error) {
 	return nil
 }
 
+// EnableSnapServices enables all services of the snap; the main use case for this is
+// the first boot of a pre-seeded image with service files already in place but not enabled.
+// XXX: it should go away once services are fixed and enabled on start.
+func EnableSnapServices(s *snap.Info, inter interacter) (err error) {
+	sysd := systemd.New(dirs.GlobalRootDir, systemd.SystemMode, inter)
+	for _, app := range s.Apps {
+		if app.IsService() {
+			svcName := app.ServiceName()
+			if err := sysd.Enable(svcName); err != nil {
+				return err
+			}
+		}
+	}
+	return nil
+}
+
 // StopServices stops service units for the applications from the snap which are services.
 func StopServices(apps []*snap.AppInfo, reason snap.ServiceStopReason, inter interacter, tm timings.Measurer) error {
 	sysd := systemd.New(dirs.GlobalRootDir, systemd.SystemMode, inter)
@@ -358,8 +399,13 @@ func ServicesEnableState(s *snap.Info, inter interacter) (map[string]bool, error
 	return snapSvcsState, nil
 }
 
-// RemoveSnapServices disables and removes service units for the applications from the snap which are services.
+// RemoveSnapServices disables and removes service units for the applications
+// from the snap which are services. The optional flag indicates whether
+// services are removed as part of undoing of first install of a given snap.
 func RemoveSnapServices(s *snap.Info, inter interacter) error {
+	if s.GetType() == snap.TypeSnapd {
+		return fmt.Errorf("internal error: removing explicit services for snapd snap is unexpected")
+	}
 	sysd := systemd.New(dirs.GlobalRootDir, systemd.SystemMode, inter)
 	nservices := 0
 
@@ -440,6 +486,7 @@ Before={{ stringsJoin .Before " "}}
 X-Snappy=yes
 
 [Service]
+EnvironmentFile=-/etc/environment
 ExecStart={{.App.LauncherCommand}}
 SyslogIdentifier={{.App.Snap.InstanceName}}.{{.App.Name}}
 Restart={{.Restart}}
@@ -687,41 +734,160 @@ func makeAbbrevWeekdays(start time.Weekday, end time.Weekday) []string {
 	return out
 }
 
+// daysRange generates a string representing a continuous range between given
+// day numbers, which due to compatiblilty with old systemd version uses a
+// verbose syntax of x,y,z instead of x..z
+func daysRange(start, end uint) string {
+	var buf bytes.Buffer
+	for i := start; i <= end; i++ {
+		buf.WriteString(strconv.FormatInt(int64(i), 10))
+		if i < end {
+			buf.WriteRune(',')
+		}
+	}
+	return buf.String()
+}
+
 // generateOnCalendarSchedules converts a schedule into OnCalendar schedules
 // suitable for use in systemd *.timer units using systemd.time(7)
 // https://www.freedesktop.org/software/systemd/man/systemd.time.html
+// XXX: old systemd versions do not support x..y ranges
 func generateOnCalendarSchedules(schedule []*timeutil.Schedule) []string {
 	calendarEvents := make([]string, 0, len(schedule))
 	for _, sched := range schedule {
 		days := make([]string, 0, len(sched.WeekSpans))
 		for _, week := range sched.WeekSpans {
 			abbrev := strings.Join(makeAbbrevWeekdays(week.Start.Weekday, week.End.Weekday), ",")
-			switch week.Start.Pos {
-			case timeutil.EveryWeek:
+
+			if week.Start.Pos == timeutil.EveryWeek && week.End.Pos == timeutil.EveryWeek {
 				// eg: mon, mon-fri, fri-mon
 				days = append(days, fmt.Sprintf("%s *-*-*", abbrev))
-			case timeutil.LastWeek:
-				// eg: mon5
-				days = append(days, fmt.Sprintf("%s *-*~7/1", abbrev))
-			default:
-				// eg: mon1, fri1, mon1-tue2
-				startDay := (week.Start.Pos-1)*7 + 1
-				endDay := week.End.Pos * 7
+				continue
+			}
+			// examples:
+			// mon1 - Mon *-*-1..7 (Monday during the first 7 days)
+			// fri1 - Fri *-*-1..7 (Friday during the first 7 days)
 
-				// NOTE: schedule mon1-tue2 (all weekdays
-				// between the first Monday of the month, until
-				// the second Tuesday of the month) is not
-				// translatable to systemd.time(7) format, for
-				// this assume all weekdays and allow the runner
-				// to do the filtering
-				if week.Start != week.End {
+			// entries below will make systemd timer expire more
+			// frequently than the schedule suggests, however snap
+			// runner evaluates current time and gates the actual
+			// action
+			//
+			// mon1-tue - *-*-1..7 *-*-8 (anchored at first
+			// Monday; Monday happens during the 7 days,
+			// Tuesday can possibly happen on the 8th day if
+			// the month started on Tuesday)
+			//
+			// mon-tue1 - *-*~1 *-*-1..7 (anchored at first
+			// Tuesday; matching Monday can happen on the
+			// last day of previous month if Tuesday is the
+			// 1st)
+			//
+			// mon5-tue - *-*~1..7 *-*-1 (anchored at last
+			// Monday, the matching Tuesday can still happen
+			// within the last 7 days, or on the 1st of the
+			// next month)
+			//
+			// fri4-mon - *-*-22-31 *-*-1..7 (anchored at 4th
+			// Friday, can span onto the next month, extreme case in
+			// February when 28th is Friday)
+			//
+			// XXX: since old versions of systemd, eg. 229 available
+			// in 16.04 does not support x..y ranges, days need to
+			// be enumerated like so:
+			// Mon *-*-1..7 -> Mon *-*-1,2,3,4,5,6,7
+			//
+			// XXX: old systemd versions do not support the last n
+			// days syntax eg, *-*~1, thus the range needs to be
+			// generated in more verbose way like so:
+			// Mon *-*~1..7 -> Mon *-*-22,23,24,25,26,27,28,29,30,31
+			// (22-28 is the last week, but the month can have
+			// anywhere from 28 to 31 days)
+			//
+			startPos := week.Start.Pos
+			endPos := startPos
+			if !week.AnchoredAtStart() {
+				startPos = week.End.Pos
+				endPos = startPos
+			}
+			startDay := (startPos-1)*7 + 1
+			endDay := (endPos) * 7
+
+			if week.IsSingleDay() {
+				// single day, can use the 'weekday' filter
+				if startPos == timeutil.LastWeek {
+					// last week of a month, which can be
+					// 22-28 in case of February, while
+					// month can have between 28 and 31 days
 					days = append(days,
-						fmt.Sprintf("*-*-%d..%d/1", startDay, endDay))
+						fmt.Sprintf("%s *-*-%s", abbrev, daysRange(22, 31)))
 				} else {
 					days = append(days,
-						fmt.Sprintf("%s *-*-%d..%d/1", abbrev, startDay, endDay))
+						fmt.Sprintf("%s *-*-%s", abbrev, daysRange(startDay, endDay)))
+				}
+				continue
+			}
+
+			if week.AnchoredAtStart() {
+				// explore the edge cases first
+				switch startPos {
+				case timeutil.LastWeek:
+					// starts in the last week of the month and
+					// possibly spans into the first week of the
+					// next month;
+					// month can have between 28 and 31
+					// days
+					days = append(days,
+						// trailing 29-31 that are not part of a full week
+						fmt.Sprintf("*-*-%s", daysRange(29, 31)),
+						fmt.Sprintf("*-*-%s", daysRange(1, 7)))
+				case 4:
+					// a range in the 4th week can span onto
+					// the next week, which is either 28-31
+					// or in extreme case (eg. February with
+					// 28 days) 1-7 of the next month
+					days = append(days,
+						// trailing 29-31 that are not part of a full week
+						fmt.Sprintf("*-*-%s", daysRange(29, 31)),
+						fmt.Sprintf("*-*-%s", daysRange(1, 7)))
+				default:
+					// can possibly spill into the next week
+					days = append(days,
+						fmt.Sprintf("*-*-%s", daysRange(startDay+7, endDay+7)))
 				}
 
+				if startDay < 28 {
+					days = append(days,
+						fmt.Sprintf("*-*-%s", daysRange(startDay, endDay)))
+				} else {
+					// from the end of the month
+					days = append(days,
+						fmt.Sprintf("*-*-%s", daysRange(startDay-7, endDay-7)))
+				}
+			} else {
+				switch endPos {
+				case timeutil.LastWeek:
+					// month can have between 28 and 31
+					// days, add trailing 29-31 that are not
+					// part of a full week
+					days = append(days, fmt.Sprintf("*-*-%s", daysRange(29, 31)))
+				case 1:
+					// possibly spans from the last week of the
+					// previous month and ends in the first week of
+					// current month
+					days = append(days, fmt.Sprintf("*-*-%s", daysRange(22, 31)))
+				default:
+					// can possibly spill into the previous week
+					days = append(days,
+						fmt.Sprintf("*-*-%s", daysRange(startDay-7, endDay-7)))
+				}
+				if endDay < 28 {
+					days = append(days,
+						fmt.Sprintf("*-*-%s", daysRange(startDay, endDay)))
+				} else {
+					days = append(days,
+						fmt.Sprintf("*-*-%s", daysRange(startDay-7, endDay-7)))
+				}
 			}
 		}
 
@@ -748,7 +914,7 @@ func generateOnCalendarSchedules(schedule []*timeutil.Schedule) []string {
 						// directly one after another
 						length -= 5 * time.Minute
 					}
-					when = when.Add(time.Duration(rand.Int63n(int64(length))))
+					when = when.Add(randutil.RandomDuration(length))
 				}
 				if when.Hour == 24 {
 					// 24:00 for us means the other end of
